import os
import sys
import pprint
import random
import time
import tqdm
import logging
import argparse
import numpy as np

import torch
import torch.nn as nn
import torch.multiprocessing as mp
import torch.distributed as dist

import losses
import models
import datasets
import lib.utils as utils
from lib.utils import AverageMeter
from optimizer.optimizer import Optimizer
from evaluation.evaler import Evaler
from scorer.scorer import Scorer
from lib.config import cfg, cfg_from_file
import readers.caption_data as dataset
import framework.run_utils
import framework.logbase
import torch.utils.data as data
from models.masking import mask
def make_masks(feature_stacks, pad_idx):

   masks = mask(feature_stacks[:, :, 0], None, pad_idx)

   return masks

class Trainer(object):
    def __init__(self, args):
        super(Trainer, self).__init__()
        self.args = args
        if cfg.SEED > 0:
            random.seed(cfg.SEED)
            torch.manual_seed(cfg.SEED)
            torch.cuda.manual_seed_all(cfg.SEED)

        self.num_gpus = torch.cuda.device_count()
        self.distributed = self.num_gpus > 11
        if self.distributed:
            torch.cuda.set_device(args.local_rank)
            torch.distributed.init_process_group(
                backend="nccl", init_method="env://"
            )
        self.device = torch.device("cuda")

        self.rl_stage = False
        self.setup_logging()
        self.setup_network()
        self.scorer = Scorer()

    def setup_logging(self):
        self.logger = logging.getLogger(cfg.LOGGER_NAME)
        self.logger.setLevel(logging.INFO)
        if self.distributed and dist.get_rank() > 0:
            return
        
        ch = logging.StreamHandler(stream=sys.stdout)
        ch.setLevel(logging.INFO)
        formatter = logging.Formatter("[%(levelname)s: %(asctime)s] %(message)s")
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)

        if not os.path.exists(cfg.ROOT_DIR):
            os.makedirs(cfg.ROOT_DIR)
        
        fh = logging.FileHandler(os.path.join(cfg.ROOT_DIR, cfg.LOGGER_NAME + '.txt'))
        fh.setLevel(logging.INFO)
        fh.setFormatter(formatter)
        self.logger.addHandler(fh)
        
        self.logger.info('Training with config:')
        self.logger.info(pprint.pformat(cfg))

    def setup_network(self):
        a=cfg.MODEL.TYPE
        model = models.create(cfg.MODEL.TYPE)

        #if self.distributed:
            # this should be removed if we update BatchNorm stats
        #    self.model = torch.nn.parallel.DistributedDataParallel(
        #        model.to(self.device), 
        #        device_ids = [self.args.local_rank], 
        #        output_device = self.args.local_rank,
        #        broadcast_buffers = False
        #    )
        #else:
        #    self.model = torch.nn.DataParallel(model).cuda()
        self.model=model.cuda()
        #self.model=torch.nn.DataParallel(model, device_ids=[0, 1]).cuda()

        if self.args.resume > 0:
            self.model.load_state_dict(
                torch.load(self.snapshot_path("caption_model", self.args.resume),
                    map_location=lambda storage, loc: storage)
            )

        self.optim = Optimizer(self.model)
        self.xe_criterion = losses.create(cfg.LOSSES.XE_TYPE).cuda()
        self.rl_criterion = losses.create(cfg.LOSSES.RL_TYPE).cuda()
        
    def setup_loader(self, epoch):
        self.training_loader = datasets.data_loader.load_train(
            self.distributed, epoch, self.coco_set)

    def eval(self, epoch):
        if (epoch + 1) % cfg.SOLVER.TEST_INTERVAL != 0:
            return None
        if self.distributed and dist.get_rank() > 0:
            return None
            
        val_res = self.val_evaler(self.model, 'val_' + str(epoch + 1))
        self.logger.info('######## Epoch (VAL)' + str(epoch + 1) + ' ########')
        self.logger.info(str(val_res))

        test_res = self.test_evaler(self.model,'test_' + str(epoch + 1))
        self.logger.info('######## Epoch (TEST)' + str(epoch + 1) + ' ########')
        self.logger.info(str(test_res))

        val = 0
        for score_type, weight in zip(cfg.SCORER.TYPES, cfg.SCORER.WEIGHTS):
            val -= val_res[score_type] * weight
        return val

    def snapshot_path(self, name, epoch):
        snapshot_folder = os.path.join(cfg.ROOT_DIR, 'anet_origin')
        return os.path.join(snapshot_folder, name + "_" + str(epoch) + ".pth")

    def save_model(self, epoch):
        if (epoch + 1) % cfg.SOLVER.SNAPSHOT_ITERS != 0:
            return
        if self.distributed and dist.get_rank() > 0:
            return
        snapshot_folder = os.path.join(cfg.ROOT_DIR, 'anet_origin')
        if not os.path.exists(snapshot_folder):
            os.mkdir(snapshot_folder)
        torch.save(self.model.state_dict(), self.snapshot_path("caption_model", epoch+1))

    ##
    def make_kwargs(self, input_seq, target_seq, gv_feat, feats_global, feats_spatial, mask_global, mask_spatial, knowledge_fts):
        seq_mask = (input_seq > 0).type(torch.cuda.LongTensor)
        seq_mask[:,0] += 1
        seq_mask_sum = seq_mask.sum(-1)
        max_len = int(seq_mask_sum.max())
        input_seq = input_seq[:, 0:max_len].contiguous()
        target_seq = target_seq[:, 0:max_len].contiguous()

        kwargs = {
            cfg.PARAM.INPUT_SENT: input_seq,
            cfg.PARAM.TARGET_SENT: target_seq,
            cfg.PARAM.GLOBAL_FEAT: gv_feat,
            cfg.PARAM.ATT_FEATS_GLOBAL: feats_global,
            cfg.PARAM.ATT_FEATS_SPATIAL: feats_spatial,
            cfg.PARAM.ATT_FEATS_MASK_GLOBAL: mask_global,
            cfg.PARAM.ATT_FEATS_MASK_SPATIAL: mask_spatial,
            cfg.PARAM.KNOWLEDGE: knowledge_fts
        }
        return kwargs

    def scheduled_sampling(self, epoch):
        if epoch > cfg.TRAIN.SCHEDULED_SAMPLING.START:
            frac = (epoch - cfg.TRAIN.SCHEDULED_SAMPLING.START) // cfg.TRAIN.SCHEDULED_SAMPLING.INC_EVERY
            ss_prob = min(cfg.TRAIN.SCHEDULED_SAMPLING.INC_PROB * frac, cfg.TRAIN.SCHEDULED_SAMPLING.MAX_PROB)
            self.model.ss_prob = ss_prob

    def display(self, iteration, data_time, batch_time, losses, loss_info):
        if iteration % cfg.SOLVER.DISPLAY != 0:
            return
        if self.distributed and dist.get_rank() > 0:
            return
        info_str = ' (DataTime/BatchTime: {:.3}/{:.3}) losses = {:.5}'.format(data_time.avg, batch_time.avg, losses.avg)
        self.logger.info('Iteration ' + str(iteration) + info_str +', lr = ' +  str(self.optim.get_lr()))
        for name in sorted(loss_info):
            self.logger.info('  ' + name + ' = ' + str(loss_info[name]))
        data_time.reset()
        batch_time.reset()
        losses.reset()

    def forward(self, kwargs):
        if self.rl_stage == False:
            logit = self.model(**kwargs)
            loss, loss_info = self.xe_criterion(logit, kwargs[cfg.PARAM.TARGET_SENT])
        else:
            ##
            ids = kwargs[cfg.PARAM.TARGET_SENT]
            gv_feat = kwargs[cfg.PARAM.GLOBAL_FEAT]
            att_feats_global = kwargs[cfg.PARAM.ATT_FEATS_GLOBAL]
            att_feats_spatial = kwargs[cfg.PARAM.ATT_FEATS_SPATIAL]
            att_mask_global = kwargs[cfg.PARAM.ATT_FEATS_MASK_GLOBAL]
            att_mask_spatial = kwargs[cfg.PARAM.ATT_FEATS_MASK_SPATIAL]

            # max
            kwargs['BEAM_SIZE'] = 1
            kwargs['GREEDY_DECODE'] = True
            kwargs[cfg.PARAM.GLOBAL_FEAT] = gv_feat
            kwargs[cfg.PARAM.ATT_FEATS_GLOBAL] = att_feats_global
            kwargs[cfg.PARAM.ATT_FEATS_SPATIAL] = att_feats_spatial
            kwargs[cfg.PARAM.ATT_FEATS_MASK_GLOBAL] = att_mask_global
            kwargs[cfg.PARAM.ATT_FEATS_MASK_SPATIAL] = att_mask_spatial

            self.model.eval()
            with torch.no_grad():
                #seq_max, logP_max = self.model.module.decode(**kwargs)
                seq_max, logP_max = self.model.decode(**kwargs)

            self.model.train()
            rewards_max, rewards_info_max = self.scorer(ids.cpu().numpy().tolist(), seq_max.data.cpu().numpy().tolist())
            #rewards_max, rewards_info_max = self.scorer(ids.data.cpu().numpy().tolist(), seq_max.data.cpu().numpy().tolist())
            #rewards_max = utils.expand_numpy(rewards_max)

            #ids = utils.expand_numpy(ids)
            #gv_feat = utils.expand_tensor(gv_feat, cfg.DATA_LOADER.SEQ_PER_IMG)
            #att_feats = utils.expand_tensor(att_feats, cfg.DATA_LOADER.SEQ_PER_IMG)
            #att_mask = utils.expand_tensor(att_mask, cfg.DATA_LOADER.SEQ_PER_IMG)

            # sample
            kwargs['BEAM_SIZE'] = 1
            kwargs['GREEDY_DECODE'] = False
            kwargs[cfg.PARAM.GLOBAL_FEAT] = gv_feat
            kwargs[cfg.PARAM.GLOBAL_FEAT] = gv_feat
            kwargs[cfg.PARAM.ATT_FEATS_GLOBAL] = att_feats_global
            kwargs[cfg.PARAM.ATT_FEATS_SPATIAL] = att_feats_spatial
            kwargs[cfg.PARAM.ATT_FEATS_MASK_GLOBAL] = att_mask_global
            kwargs[cfg.PARAM.ATT_FEATS_MASK_SPATIAL] = att_mask_spatial

            #seq_sample, logP_sample = self.model.module.decode(**kwargs)
            seq_sample, logP_sample = self.model.decode(**kwargs)
            #rewards_sample, rewards_info_sample = self.scorer(ids, seq_sample.data.cpu().numpy().tolist())
            rewards_sample, rewards_info_sample = self.scorer(ids.data.cpu().numpy().tolist(), seq_sample.data.cpu().numpy().tolist())

            rewards = rewards_sample - rewards_max
            rewards = torch.from_numpy(rewards).float().cuda()
            loss = self.rl_criterion(seq_sample, logP_sample, rewards)
            
            loss_info = {}
            for key in rewards_info_sample:
                loss_info[key + '_sample'] = rewards_info_sample[key]
            for key in rewards_info_max:
                loss_info[key + '_max'] = rewards_info_max[key]

        return loss, loss_info

    def train(self):
        self.model.train()
        self.optim.zero_grad()
        path_cfg_file = './paths/path_anet.json'
        path_cfg = framework.run_utils.gen_common_pathcfg(path_cfg_file, is_train=True)
        trn_data = dataset.CaptionDataset(path_cfg.name_file['trn'], path_cfg.ft_root, path_cfg.cap_file, path_cfg.word2int_file, path_cfg.int2word_file, path_cfg.knowledge_file['trn'], 150, is_train=True, _logger=None)
        train_loader = data.DataLoader(trn_data, batch_size=20, shuffle=True, num_workers=4)

        iteration = 0
        for epoch in  range(cfg.SOLVER.MAX_EPOCH):
            if epoch == cfg.TRAIN.REINFORCEMENT.START:
                self.rl_stage = True
            #self.setup_loader(epoch)
            start = time.time()
            data_time = AverageMeter()
            batch_time = AverageMeter()
            losses = AverageMeter()
            ##
            print(len(train_loader))
            for i, batch in enumerate(train_loader):
                data_time.update(time.time() - start)
                knowledge_fts = torch.LongTensor(batch['knowledge_feature']).cuda()
                trg = torch.LongTensor(batch['caption_ids']).cuda()
                img_fts_global = torch.FloatTensor(batch['img_ft_temporal']).cuda()
                img_fts_spatial = torch.FloatTensor(batch['img_ft_spatial']).cuda()
                ft_len = torch.LongTensor(batch['ft_len']).cuda()
                img_fts_global = img_fts_global[:,:max(ft_len)]
                img_fts_spatial = img_fts_spatial[:, :max(ft_len)]
                trg = trg[:,:max(batch['id_len'])]
                B1, num1, spatial, num2=img_fts_spatial.size()
                img_fts_spatial = img_fts_spatial.view(B1*num1, spatial, num2)
                mask_global = torch.ones((B1, num1), dtype=torch.long).cuda()
                mask_spatial = torch.ones((B1*num1, spatial+1), dtype=torch.long).cuda()
                gv_feat = torch.zeros(B1, 1).cuda()
                [B1, seq_len] = trg.size()
                caption_idx_new = torch.zeros((B1, seq_len - 1), dtype=torch.long).cuda()
                for i in range(B1):
                    caption_tmp=trg[i, 1:]
                    mask0 = (caption_tmp == 0)
                    mask = ~((caption_tmp==1)+(caption_tmp==3))
                    caption_tmp = caption_tmp * mask
                    caption_tmp = caption_tmp + mask0
                    caption_idx_new[i]=caption_tmp
                seq0 = torch.zeros((B1, 1), dtype=torch.long).cuda()
                input_seq = torch.cat((seq0, caption_idx_new[:,:-1]), dim=1)
                target_seq = caption_idx_new
                kwargs = self.make_kwargs(input_seq, target_seq, gv_feat, img_fts_global, img_fts_spatial, mask_global, mask_spatial, knowledge_fts)
                loss, loss_info = self.forward(kwargs)
                loss.backward()
                utils.clip_gradient(self.optim.optimizer, self.model,
                    cfg.SOLVER.GRAD_CLIP_TYPE, cfg.SOLVER.GRAD_CLIP)
                self.optim.step()
                self.optim.zero_grad()
                self.optim.scheduler_step('Iter')
                
                batch_time.update(time.time() - start)
                start = time.time()
                losses.update(loss.item())
                if iteration % 20 == 0:
                    print('iteration: [%d], loss_avg: [%.4f]' %(iteration, losses.avg))
                    f1=open('./experiments/configs/anet_origin/log.txt', 'a+')
                    print('iteration: [%d], loss_avg: [%.4f]' %(iteration, losses.avg), file=f1)
                iteration += 1

                if self.distributed:
                    dist.barrier()
        
            self.save_model(epoch)
            self.scheduled_sampling(epoch)

            if self.distributed:
                dist.barrier()

def parse_args():
    """
    Parse input arguments
    """
    parser = argparse.ArgumentParser(description='Image Captioning')
    parser.add_argument('--folder', type=str, default='./experiments/configs')
    parser.add_argument('--local_rank', type=int, default=0)
    parser.add_argument('--resume', type=int, default=-1)

    args = parser.parse_args()
    return args

if __name__ == '__main__':
    args = parse_args()
    print('Called with args:')
    print(args)

    if args.folder is not None:
        cfg_from_file(os.path.join(args.folder, 'config.yml'))
    cfg.ROOT_DIR = args.folder

    trainer = Trainer(args)
    trainer.train()
